#ifndef SPI_MISC_H__
#define SPI_MISC_H__

//-----------------------------------------------------------------------------------------------------------
/* 
	ėpNXE֐

	last update: 2005/09/8
*/
//-----------------------------------------------------------------------------------------------------------

/*
	// Debug function

	// ėpPointer
	union MultiPointer;

	// File class
	class AutoFile;	
	class MapFile;

	// Heap class
	class AutoHeap;
	class LocalHeap;

	// Endianϊ : big to little
	WORD GetWordBE(LPVOID src);
	DWORD GetDwordBE(LPVOID src);

	// Path Utility

	// Image Utility

*/


#ifndef	_T
#	ifdef _UNICODE
#		define _T(x)	L ## x
#else
#		define _T(x)	x
#	endif
#endif

/* Debug function */
#ifdef _DEBUG
inline void PrintInt(int c, LPSTR str = NULL)
{
	char buf[20];
	wsprintf(buf, "%d", c);
	if (str == NULL) MessageBox(NULL, buf, "Print", 0);
	else MessageBox(NULL, buf, str, 0);
}

inline void PrintHex(int c, LPSTR str = NULL)
{
	char buf[20];
	wsprintf(buf, "%x", c);
	if (str == NULL) MessageBox(NULL, buf, "Print", 0);
	else MessageBox(NULL, buf, str, 0);
}

#ifndef	SpiAssert
inline void SpiTrace(LPSTR format, ...)
{
	TCHAR msg[MAX_PATH];
	va_list argList;
	va_start(argList, format);
	wvsprintf(msg, format, argList);
	va_end(argList);
	OutputDebugString(msg);
}

#define SpiAssert(exp){		\
if (!(exp)){		\
		SpiTrace(_T("Assertion failed! in %s (%d) %s\n"),	\
            _T(__FILE__), __LINE__, _T(#exp));			\
		DebugBreak();	\
	}		\
}
#endif	// SpiAssert

#else	// _DEBUG
inline void PrintInt(int c, LPSTR str = NULL) { c; str;}
inline void PrintHex(int c, LPSTR str = NULL) { c; str;}
inline void SpiTrace(LPSTR format, ...) { format; }
#ifndef	SpiAssert
#define SpiAssert(exp)
#endif

#endif	// _DEBUG

inline void MsgBoxF(LPSTR title, LPSTR format, ...)
{
	TCHAR msg[MAX_PATH];
	va_list argList;
	va_start(argList, format);
	wvsprintf(msg, format, argList);
	va_end(argList);
	MessageBox(NULL, msg, title, 0);
}

inline DWORD WriteData(LPCSTR filename, void* buff, DWORD size, bool bAppend = false)
{
	DWORD tmp	= 0;
	DWORD createDisposition	= bAppend ? OPEN_ALWAYS : CREATE_ALWAYS;
	HANDLE hFile = CreateFile(filename, GENERIC_WRITE, FILE_SHARE_READ, NULL, createDisposition, 0, NULL);
	if (hFile != INVALID_HANDLE_VALUE)
	{
		if (bAppend)
			SetFilePointer(hFile, 0, NULL, FILE_END);
		WriteFile(hFile, buff, size, &tmp, NULL);
		CloseHandle(hFile);
	}
	return tmp;
}

// API G[bZ[W\
inline void ApiError(LPTSTR title)
{
	char buf[256];
	
	::FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, 0, GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), buf, 256, NULL);
	::MessageBox(NULL, buf, title, MB_OK);
}

/* ----------------------------------------------------------------------------------------------------------- */
/* GfBAϊ */
inline WORD GetWordBE(void* src)
{
	BYTE* buf	= reinterpret_cast<BYTE*>(src);
	WORD wd	= (buf[0] << 8) + buf[1];

	return wd;
}

inline DWORD GetDwordBE(void* src)
{
	BYTE* buf	= reinterpret_cast<BYTE*>(src);
	DWORD dw	= (buf[0] << 8) + buf[1];
	dw			= (dw << 8) + buf[2];
	dw			= (dw << 8) + buf[3];
	return dw;
}

/* ----------------------------------------------------------------------------------------------------------- */
/* RGB color */
class CRgb : public tagRGBQUAD
{
public:
	CRgb()
	{ }

	CRgb(BYTE red, BYTE green, BYTE blue, BYTE alpha = 0)
	{
		rgbBlue		= blue;
		rgbGreen	= green;
		rgbRed		= red;
		rgbReserved	= alpha;
	}

	CRgb(RGBQUAD rgb)
	{
		*(RGBQUAD*)this	= rgb; 
	}

	CRgb(COLORREF color)
	{
		rgbRed		= static_cast<BYTE>(color);
		rgbGreen	= static_cast<BYTE>(color >> 8);
		rgbBlue		= static_cast<BYTE>(color >> 16);
		rgbReserved	= 0;
	}

	COLORREF GetColorRef() const
	{
		return RGB(rgbRed, rgbGreen, rgbBlue);
	}

	operator COLORREF() const
	{
		return GetColorRef();
	}
};


/* ----------------------------------------------------------------------------------------------------------- */
/* Smart File */
// X}[g|C^݂ɎŃt@C
class AutoFile
{
protected:
	HANDLE m_hFile;
	AutoFile(const AutoFile &file);		// ֎~
	const AutoFile& operator =(const AutoFile& file);

public:
	// \zE
	AutoFile(HANDLE hFile = INVALID_HANDLE_VALUE)
		: m_hFile(hFile)
	{}

	~AutoFile()
	{
		Close();
	}

	// Agr[g
	HANDLE GetHandle() const { return m_hFile; }

	// Iy[V
	HANDLE Open(LPCTSTR FileName, bool bReadOnly = true)
	{
		if (bReadOnly)
		{
			m_hFile = CreateFile(FileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		} else {
			m_hFile = CreateFile(FileName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
		}
		return m_hFile;
	}

	HANDLE Open(LPCTSTR FileName, bool bReadOnly, DWORD openFlags)
	{
		if (bReadOnly)
		{
			m_hFile = CreateFile(FileName, GENERIC_READ, FILE_SHARE_READ, NULL, openFlags, 0, NULL);
		} else {
			m_hFile = CreateFile(FileName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, openFlags, 0, NULL);
		}
		return m_hFile;
	}

	HANDLE OpenRead(LPCTSTR FileName, ULONG shareMode = FILE_SHARE_READ, ULONG openFlags = OPEN_EXISTING, ULONG attributes = FILE_ATTRIBUTE_NORMAL)
	{
		m_hFile = CreateFile(FileName, GENERIC_READ, shareMode, NULL, openFlags, attributes, NULL);
		return m_hFile;
	}

	HANDLE OpenWrite(LPCTSTR FileName, ULONG shareMode = 0, ULONG openFlags = CREATE_ALWAYS, ULONG attributes = FILE_ATTRIBUTE_NORMAL)
	{
		m_hFile = CreateFile(FileName, GENERIC_WRITE, shareMode, NULL, openFlags, attributes, NULL);
		return m_hFile;
	}

	HANDLE ReOpen(LPCTSTR FileName, bool bReadOnly = true)
	{
		CloseHandle(m_hFile);

		if (bReadOnly)
			m_hFile = CreateFile(FileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		else
			m_hFile = CreateFile(FileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);

		return m_hFile;
	}

	void Close()
	{
		CloseHandle(m_hFile);
		m_hFile	= INVALID_HANDLE_VALUE;
	}

	void Attach(HANDLE hFile)
	{
		m_hFile	= hFile;
	}

	HANDLE Detach()
	{
		HANDLE hFile	= m_hFile;
		m_hFile		= INVALID_HANDLE_VALUE;

		return hFile;
	}

	DWORD Seek(long lDistance, DWORD flag = FILE_CURRENT, LONG* hDistance = NULL)
	{
		return SetFilePointer(m_hFile, lDistance, hDistance, flag);
	}

	BOOL Read(void* dest, DWORD size, DWORD* lpReadBytes)
	{
		return ReadFile(m_hFile, dest, size, lpReadBytes, NULL);
	}

	BOOL Write(const void* src, DWORD size, DWORD* lpWriteBytes)
	{
		return WriteFile(m_hFile, src, size, lpWriteBytes, NULL);
	}

	DWORD GetFileSize(DWORD* lpFileSizeHigh = NULL)
	{
		return ::GetFileSize(m_hFile, lpFileSizeHigh);
	}

	BOOL SetEndOfFile()
	{
		return ::SetEndOfFile(m_hFile);
	}
};

/* ----------------------------------------------------------------------------------------------------------- */
/* Mapping File */
class MapFile
{
private:
	HANDLE m_hFile;
	HANDLE m_hMap;
	LPVOID m_MapView;
	MapFile(const MapFile& hFile);		// ֎~
	const MapFile& operator =(const MapFile& file);

public:
	// \zE
	MapFile(HANDLE hFile = INVALID_HANDLE_VALUE)
		: m_hFile(hFile), m_hMap(NULL), m_MapView(NULL)
	{}

	~MapFile()
	{
		Close();
	}

	// 
	HANDLE GetHandle() const { return m_hFile; }
	DWORD GetSize(DWORD* sizeHigh) const { return ::GetFileSize(m_hFile, sizeHigh); }

	// Iy[V
	HANDLE Open(LPCTSTR FileName, bool bReadOnly = true)
	{
		if (bReadOnly == true)
			m_hFile = CreateFile(FileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		else
			m_hFile = CreateFile(FileName, GENERIC_WRITE | GENERIC_READ, 0, NULL, OPEN_ALWAYS, 0, NULL);

		return m_hFile;
	}

	void Close()
	{
		if (m_MapView != NULL)
			UnMap();
		if (m_hFile != INVALID_HANDLE_VALUE)
		{
			CloseHandle(m_hFile);
			m_hFile	= INVALID_HANDLE_VALUE;
		}
	}

	void Attach(HANDLE hFile)
	{
		m_hFile	= hFile;
	}

	HANDLE Detach()
	{
		UnMap();

		HANDLE hFile	= m_hFile;
		m_hFile		= INVALID_HANDLE_VALUE;

		return hFile;
	}

	LPVOID Map(bool bReadOnly = true, DWORD size = 0, DWORD offsetLow = 0, DWORD offsetHigh = 0)
	{
		SYSTEM_INFO SysInfo;
		GetSystemInfo(&SysInfo);
		DWORD alignment	= offsetLow & (SysInfo.dwAllocationGranularity - 1);

		if (bReadOnly == true)
		{
			m_hMap		= CreateFileMapping(m_hFile, NULL, PAGE_READONLY, 0, size + alignment, NULL);
			if (m_hMap == NULL)
				return NULL;
			m_MapView	= MapViewOfFile(m_hMap, FILE_MAP_READ, offsetHigh, offsetLow - alignment, size + alignment);
		} else {
			m_hMap		= CreateFileMapping(m_hFile, NULL, PAGE_READWRITE, 0, size + alignment, NULL);
			if (m_hMap == NULL)
				return NULL;
			m_MapView	= MapViewOfFile(m_hMap, FILE_MAP_WRITE, offsetHigh, offsetLow - alignment, size + alignment);
		}

		return (m_MapView == NULL) ? NULL : ((BYTE*)m_MapView + alignment);
	}

	LPVOID ReMap(bool bReadOnly = true, DWORD offsetLow = 0, DWORD size = 0, DWORD offsetHigh = 0)
	{
		UnmapViewOfFile(m_MapView);

		SYSTEM_INFO SysInfo;
		GetSystemInfo(&SysInfo);
		DWORD alignment	= offsetLow & (SysInfo.dwAllocationGranularity - 1);

		if (bReadOnly == true)
			m_MapView	= MapViewOfFile(m_hMap, FILE_MAP_READ, offsetHigh, offsetLow - alignment, size + alignment);
		else
			m_MapView	= MapViewOfFile(m_hMap, FILE_MAP_WRITE, offsetHigh, offsetLow - alignment, size + alignment);

		return (m_MapView == NULL) ? NULL : ((BYTE*)m_MapView + alignment);
	}

	void UnMap()
	{
		UnmapViewOfFile(m_MapView);
		CloseHandle(m_hMap);
		m_MapView	= NULL;
	}
};

/* ----------------------------------------------------------------------------------------------------------- */
/* Auto Heap */
class AutoHeap
{
private:
	HANDLE m_hf;
	AutoHeap(const AutoHeap &obj);	// Rs[֎~
	const AutoHeap& operator =(const AutoHeap& file);

public:
	// \zE
	AutoHeap() : m_hf(NULL)
	{}

	~AutoHeap()
	{
		if (m_hf != NULL)
			Free();
	}

	// Agr[g
	void* GetData() const
	{
		return m_hf;
	}

	// Iy[V
	void* Alloc(SIZE_T size, DWORD flag = 0)
	{
		m_hf	= HeapAlloc(GetProcessHeap(), flag, size);
		return m_hf;
	}

	void* ReAlloc(SIZE_T size, DWORD flag = 0)
	{
		HANDLE hf	= HeapReAlloc(GetProcessHeap(), flag, m_hf, size);
		if (hf == NULL)
			return NULL;

		return m_hf	= hf;
	}

	void Free()
	{
		HeapFree(GetProcessHeap(), 0, m_hf);
		m_hf	= NULL;
	}

	void* Attach(void*  hf)
	{
		if (m_hf != NULL)
			Free();
		return m_hf	= hf;
	}

	void*  Detach()
	{
		HANDLE hf	= m_hf;
		m_hf	= NULL;
		return hf;
	}
};

/* ----------------------------------------------------------------------------------------------------------- */
/* Auto Local Heap */
class LocalHeap
{
private:
	HLOCAL m_hLocal;
	LocalHeap(const LocalHeap& obj);	// Rs[֎~
	const LocalHeap& operator = (const LocalHeap& other);

public:
	// \zE
	LocalHeap() : m_hLocal(NULL)
	{}

	~LocalHeap()
	{
		if (m_hLocal != NULL)
			Free();
	}

	// Agr[g
	void* GetData() const
	{
		if (m_hLocal == NULL)
			return NULL;
		return LocalLock(m_hLocal);
	}

	// Iy[V
	void* Alloc(SIZE_T size, DWORD flag = LHND)
	{
		m_hLocal	= ::LocalAlloc(flag, size);
		if (flag & LMEM_MOVEABLE)
			return ::LocalLock(m_hLocal);
		else
			return m_hLocal;
	}

	void* ReAlloc(SIZE_T size, DWORD flag = LHND)
	{
		HLOCAL hLocal	= ::LocalReAlloc(m_hLocal, size, flag);
		if (hLocal == NULL)
			return NULL;
		m_hLocal	= hLocal;
		if (flag & LMEM_MOVEABLE)
			return ::LocalLock(m_hLocal);
		else
			return m_hLocal;
	}

	void Free()
	{
		::LocalFree(m_hLocal);
		m_hLocal = NULL;
	}

	void* Attach(HLOCAL hLocal)
	{
		if (m_hLocal != NULL)
			Free();

		m_hLocal	= hLocal;
		return ::LocalLock(m_hLocal);
	}

	HLOCAL Detach()
	{
		HLOCAL hLocal	= m_hLocal;
		::LocalUnlock(m_hLocal);
		m_hLocal	= NULL;
		return hLocal;
	}
};

/* ----------------------------------------------------------------------------------------------------------- */
/* LXǧx} */
template <class TDest>
class CReinterpretCast
{
public:
	template <class TSrc>
	TDest operator ()(TSrc src)
	{
		union U_CONVERT
		{
			TDest dest;
			TSrc src;
		};
		U_CONVERT u;
		u.src	= src;
		return u.dest;
	}
};

/* ----------------------------------------------------------------------------------------------------------- */
/* Multi Pointer */

union MultiPointer
{
	BYTE* lpByte;
	WORD* lpWord;
	short* lpShort;
	DWORD* lpDword;
	LPSTR lpStr;
};

/* ----------------------------------------------------------------------------------------------------------- */
/* image utility */

// src㉺]destɏo(4 byte alignment)
inline bool DIBInvert(BYTE* dest, const BYTE* src, const BITMAPINFOHEADER& rInfo)
{
	WORD ColorBytes	= rInfo.biBitCount >> 3;
	DWORD RealBytes	= rInfo.biWidth * ColorBytes;	// 1C̎ۂ̃TCY
	DWORD LineBytes	= (RealBytes + 3) & (~3);		// 1C̃_~[܂ރTCY
	src	+= RealBytes * (rInfo.biHeight - 1);

	for (long i = 0;i < rInfo.biHeight;i++)
	{
		memcpy(dest, src, RealBytes);
		dest	+= LineBytes;
		src		-= RealBytes;
	}
	return true;
}

// PCSwapď㉺]
inline bool DIBInvert(BYTE* pTop, const BITMAPINFOHEADER& rInfo)
{
	WORD ColorBytes	= rInfo.biBitCount >> 3;
	DWORD LineBytes	= rInfo.biWidth * ColorBytes;	// 1C̎ۂ̃TCY
	BYTE* pBottom	= &pTop[LineBytes * (rInfo.biHeight - 1)];	// ԍŌ̃C̐擪
	AutoHeap TmpHeap;
	BYTE* pLineBuf	= (BYTE*)TmpHeap.Alloc(LineBytes);
	if (pLineBuf == NULL)
		return false;

	for (long i = 0;i < rInfo.biHeight / 2;i++)
	{
		// PCSwap
		memcpy(pLineBuf, pTop, LineBytes);
		memcpy(pTop, pBottom, LineBytes);
		memcpy(pBottom, pLineBuf, LineBytes);
		pTop	+= LineBytes;
		pBottom	-= LineBytes;
	}
	return true;
}

// Color mask
inline BYTE RGB555_GetRed(WORD pixel)
{
	return static_cast<BYTE>((pixel & 0x7C00) >> 7);
}

inline BYTE RGB555_GetGreen(WORD pixel)
{
	return static_cast<BYTE>((pixel & 0x03E0) >> 2);
}

inline BYTE RGB555_GetBlue(WORD pixel)
{
	return static_cast<BYTE>((pixel & 0x001F) << 3);
}

// AlphaBlend
inline BYTE AlphaBlend(BYTE Data1, BYTE Data2, BYTE Alpha)
{
	return static_cast<BYTE>((Data1 - Data2) * (Alpha + 1) >> 8) + Data2;
}

inline BYTE AlphaBlend(BYTE Data, BYTE Alpha)
{
	return static_cast<BYTE>(Data * (Alpha + 1) >> 8);
}

inline void AlphaBlendPixel(void* _dest, const void* _src)
{
	BYTE* dest	= reinterpret_cast<BYTE*>(_dest);
	const BYTE* src	= reinterpret_cast<const BYTE*>(_src);
	if (src[3] == 0xFF)
	{
		dest[0]	= src[0];
		dest[1]	= src[1];
		dest[2]	= src[2];

	} else if (src[3] == 0x00)
	{
		dest[0]	= 0x00;
		dest[1]	= 0x00;
		dest[2]	= 0x00;
	} else {
		dest[0]	= AlphaBlend(src[0], src[3]);
		dest[1]	= AlphaBlend(src[1], src[3]);
		dest[2]	= AlphaBlend(src[2], src[3]);
	}
}

inline void AlphaBlendPixel(void* _dest, const void* _src, RGBQUAD bgColor)
{
	BYTE* dest	= reinterpret_cast<BYTE*>(_dest);
	const BYTE* src	= reinterpret_cast<const BYTE*>(_src);
	if (src[3] == 0xFF)
	{
		dest[0]	= src[0];
		dest[1]	= src[1];
		dest[2]	= src[2];
	} else if (src[3] == 0x00)
	{
		dest[0]	= bgColor.rgbBlue;
		dest[1]	= bgColor.rgbGreen;
		dest[2]	= bgColor.rgbRed;
	} else {
		dest[0]	= AlphaBlend(src[0], bgColor.rgbBlue, src[3]);
		dest[1]	= AlphaBlend(src[1], bgColor.rgbGreen, src[3]);
		dest[2]	= AlphaBlend(src[2], bgColor.rgbRed, src[3]);
	}
}

inline DWORD GetPaletteSize(const BITMAPINFOHEADER& bmih)
{
	return bmih.biClrUsed ? (bmih.biClrUsed << 2) : (4 << bmih.biBitCount);
}

inline DWORD GetRowBytes(const BITMAPINFOHEADER& bmih)
{
	return ((bmih.biWidth * bmih.biBitCount >> 3) + 3) & ~3;
}

inline DWORD GetDIBSize(const BITMAPINFOHEADER& bmih)
{
	return GetRowBytes(bmih) * bmih.biHeight;
}

// return WIN32 heap
// 32bit̓tƂč
inline void* DIBConvert24(const BITMAPINFO& bmi, const void* pSrc, COLORREF bgColor = ~0)
{
	// allocate memory
	const BITMAPINFOHEADER& bmih	= bmi.bmiHeader;
	DWORD rowBytes	= (bmih.biWidth * 3 + 3) & ~3;
	DWORD imageSize	= rowBytes * bmih.biHeight;
	BYTE* pDest	= reinterpret_cast<BYTE*>(HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, imageSize));
	if (pDest == NULL)
		return NULL;

	// convert
	BYTE* pbyDest	= pDest;
	BYTE* pbySrc	= (BYTE*)pSrc;
	if (bmih.biBitCount == 8)
	{
		DWORD srcRowBytes	= (bmih.biWidth + 3) & ~3;
		for (long i = 0;i < bmih.biHeight;i++)
		{
			BYTE* dest	= pbyDest;
			BYTE* src	= pbySrc;
			for (long j = 0;j < bmih.biWidth;j++)
			{
				BYTE index	= *src++;
				*dest++	= bmi.bmiColors[index].rgbBlue;
				*dest++	= bmi.bmiColors[index].rgbGreen;
				*dest++	= bmi.bmiColors[index].rgbRed;
			}
			pbyDest	= &pbyDest[rowBytes];
			pbySrc	= &pbySrc[srcRowBytes];
		}
	}
	else if(bmih.biBitCount == 32)
	{
		if (bgColor > 0x00FFFFFF)
		{
			// disable alpha
			for (long i = 0;i < bmih.biHeight;i++)
			{
				BYTE* pPixel	= pbyDest;
				for (long j = 0;j < bmih.biWidth;j++)
				{
					*pPixel++	= pbySrc[0];
					*pPixel++	= pbySrc[1];
					*pPixel++	= pbySrc[2];
					pbySrc	= &pbySrc[4];
				}
				pbyDest	= &pbyDest[rowBytes];
			}
		}
		else if (bgColor == 0x00)
		{
			// alpha = 0
			for (long i = 0;i < bmih.biHeight;i++)
			{
				BYTE* pPixel	= pbyDest;
				for (long j = 0;j < bmih.biWidth;j++)
				{
					AlphaBlendPixel(pPixel, pbySrc);
					pbySrc	= &pbySrc[4];
					pPixel	= &pPixel[3];
				}
				pbyDest	= &pbyDest[rowBytes];
			}
		} else {
			// alpha != 0
			CRgb rgb(bgColor);
			for (long i = 0;i < bmih.biHeight;i++)
			{
				BYTE* pPixel	= pbyDest;
				for (long j = 0;j < bmih.biWidth;j++)
				{
					AlphaBlendPixel(pPixel, pbySrc, rgb);
					pbySrc	= &pbySrc[4];
					pPixel	= &pPixel[3];
				}
				pbyDest	= &pbyDest[rowBytes];
			}
		}
	}
	else if(bmih.biBitCount == 16)
	{
		DWORD srcRowBytes	= (bmih.biWidth * 2 + 3) & ~3;
		for (long i = 0;i < bmih.biHeight;i++)
		{
			BYTE* dest	= pbyDest;
			WORD* src	= (WORD*)pbySrc;
			for (long j = 0;j < bmih.biWidth;j++, src++)
			{
				*dest++	= RGB555_GetBlue(*src);
				*dest++	= RGB555_GetGreen(*src);
				*dest++	= RGB555_GetRed(*src);
			}
			pbyDest	= &pbyDest[rowBytes];
			pbySrc	= &pbySrc[srcRowBytes];
		}
	}
	else if(bmih.biBitCount == 4)
	{
		DWORD srcRowBytes	= ((bmih.biWidth >> 1) + 3) & ~3;
		for (long i = 0;i < bmih.biHeight;i++)
		{
			BYTE* dest	= pbyDest;
			BYTE* src	= pbySrc;
			for (long j = 0;j < bmih.biWidth;j++, src++)
			{
				BYTE index	= *src >> 4;
				*dest++	= bmi.bmiColors[index].rgbBlue;
				*dest++	= bmi.bmiColors[index].rgbGreen;
				*dest++	= bmi.bmiColors[index].rgbRed;
				if (++j >= bmih.biWidth)
					break;

				index	= *src & 0x0F;
				*dest++	= bmi.bmiColors[index].rgbBlue;
				*dest++	= bmi.bmiColors[index].rgbGreen;
				*dest++	= bmi.bmiColors[index].rgbRed;
			}
			pbyDest	= &pbyDest[rowBytes];
			pbySrc	= &pbySrc[srcRowBytes];
		}
	}
	else if(bmih.biBitCount == 1)
	{
		DWORD srcRowBytes	= ((bmih.biWidth >> 3) + 3) & ~3;
		for (long i = 0;i < bmih.biHeight;i++)
		{
			BYTE* dest	= pbyDest;
			BYTE* src	= pbySrc;
			for (long j = 0;j < bmih.biWidth;)
			{
				BYTE index8	= *src++;
				for (BYTE m = 0;m < 8 && j < bmih.biWidth;m++, j++)
				{
					BYTE index	= (0x80 & index8) >> 7;	// MSBo
					index8	<<= 1;
					*dest++	= bmi.bmiColors[index].rgbBlue;
					*dest++	= bmi.bmiColors[index].rgbGreen;
					*dest++	= bmi.bmiColors[index].rgbRed;
				}
			}
			pbyDest	= &pbyDest[rowBytes];
			pbySrc	= &pbySrc[srcRowBytes];
		}
	} else {
		HeapFree(GetProcessHeap, 0, pDest);
		pDest	= NULL;
	}
	
	return pDest;
}


/* ----------------------------------------------------------------------------------------------------------- */
/* Path Utility */
namespace Path
{

inline bool IsDBCSLeadByte(BYTE ch)
{
	if (ch >= 0x81)
	{
		if (ch <= 0x9F || (0xE0 <= ch && ch <= 0xFC))
			return true;
	}
	return false;
}

inline LPTSTR FindFileName(LPCTSTR path)
{
	LPCTSTR ret	= path;
	for (;*path != '\0';path++)
	{
		if (*path == '\\')
		{
			ret	= &path[1];
		}
		else if (Path::IsDBCSLeadByte(*path))
		{
			path++;
		}
	}
	return const_cast<LPTSTR>(ret);
}

inline LPTSTR RemoveFileSpec(LPTSTR path)
{
	LPTSTR ret	= NULL;
	for (;*path != '\0';path++)
	{
		if (path[0] == '\\' && path[1] != '\0')
		{
			ret		=  path;
		}
		else if (Path::IsDBCSLeadByte(*path))
		{
			path++;
		}
	}
	if (ret != NULL)
		*ret	= '\0';
	return ret;
}

inline LPTSTR FindExtension(LPCTSTR path)
{
	LPCTSTR ret	= path;
	for (int i = 0;path[i] != '\0';i++)
	{
		if (path[i]	== '.')
			ret	= &path[i+1];
	}
	return const_cast<LPTSTR>(ret);
}

inline void  RemoveExtension(LPTSTR path)
{
	LPTSTR ext_pos	= FindExtension(path);
	if (ext_pos[-1] == '.'){
		ext_pos[-1]	= '\0';
	}
}

// root̎qȂA΃pX
inline LPCTSTR GetRelative(LPCTSTR root, LPCTSTR path)
{
	TCHAR fileName[MAX_PATH + 1];
	int length	= lstrlen(root) + 1;
	lstrcpyn(fileName, path, length);
	if (!lstrcmpi(root, fileName))
	{
		path	= &path[length];
	}
	return path;
}

#ifdef	_INC_SHLWAPI
inline BOOL RelativePathTo(LPTSTR dest, LPCTSTR root, LPCTSTR src)
{
	int length	= lstrlen(root) + 1;
	lstrcpyn(dest, src, length);

	if (lstrcmpi(root, dest))
	{
		return ::PathRelativePathTo(dest, root, FILE_ATTRIBUTE_DIRECTORY, src, FILE_ATTRIBUTE_NORMAL);
	}
	lstrcpy(dest, &src[length]);
	return TRUE;
}
#endif

inline int AddBackSlash(LPTSTR path)
{
	int len	= lstrlen(path);
	if (len > 0 && path[len - 1] != '\\' && path[len - 1] != '/')
	{
		path[len++]	= '\\';
		path[len]	= '\0';
	}
	return len;
}

inline int RemoveBackSlash(LPTSTR path)
{
	int len	= lstrlen(path);
	if (len > 0 && path[len - 1] == '\\' || path[len - 1] == '/')
	{
		path[--len]	= '\0';
	}
	return len;
}

inline void Combine(LPTSTR dest, LPCTSTR fileName, int maxLength)
{
	int pos	= AddBackSlash(dest);
	lstrcpyn(&dest[pos], fileName, maxLength - pos);
}

inline void  RenameExtension(LPTSTR path, LPCTSTR ext)
{
	LPTSTR ext_pos	= FindExtension(path);
	if (ext_pos[-1] != '.'){
		*ext_pos++	= '.';
		*ext_pos	= '\0';
	}
	lstrcpy(ext_pos, ext);
}

inline bool FileExists(LPCTSTR path)
{
	DWORD dwAttributes	= ::GetFileAttributes(path);
	return dwAttributes != 0xFFFFFFFF;
}

inline bool IsDirectory(LPCTSTR path)
{
	DWORD dwAttributes	= ::GetFileAttributes(path);
	if (dwAttributes == 0xFFFFFFFF)
		return false;

	if (FILE_ATTRIBUTE_DIRECTORY & dwAttributes)
		return true;
	return false;
}

inline bool IsRelative(LPCTSTR path)
{
	return path[1] != ':';
}

template <class T>
inline void AddQuote(T& dest, LPCTSTR src)
{
	if (src[0] != '\"')
	{
		dest	= '\"';
		dest	+= src;
		dest	+= '\"';
	} else dest	= src;
}

// internal
inline bool _MakeSureDirectoryExists(LPTSTR path)	// ċAp
{
	if (! ::CreateDirectory(path, NULL))
	{
		DWORD createErr	= ::GetLastError();
		if (createErr == ERROR_PATH_NOT_FOUND)
		{
			// ȂƂ͂߂i[vȂj
			LPTSTR delimiter	= RemoveFileSpec(path);
			if (!delimiter || !_MakeSureDirectoryExists(path))
				return false;
			*delimiter	= '\\';
			if (!::CreateDirectory(path, NULL))	// Ē
				return false;
		} else if (createErr != ERROR_ALREADY_EXISTS)
			return false;
	}
	return true;
}

inline bool CreateDirectory(LPCTSTR path, bool bOverWrite = true)
{
	if (! ::CreateDirectory(path, NULL))
	{
		DWORD createErr	= ::GetLastError();
		if (createErr == ERROR_PATH_NOT_FOUND)
		{
			TCHAR parent[MAX_PATH + 1];
			lstrcpyn(parent, path, MAX_PATH);
			if (!RemoveFileSpec(parent) || !_MakeSureDirectoryExists(parent))
				return false;
			if (!::CreateDirectory(path, NULL))	// Ē
				return false;
		}
		else if (createErr == ERROR_ALREADY_EXISTS)
		{
			return bOverWrite;
		} else return false;
	}
	return true;
}

};	// namespace Path

/* ----------------------------------------------------------------------------------------------------------- */
#endif // SPI_MISC_H__